// $Id$
//
// Copyright (C) 2010
//     Stefan Bund <info@j32.de>

/** \file
    \brief PythonPublisher non-inline non-template implementation */

#include "PythonPublisher.hh"
//#include "PythonPublisher.ih"

// Custom includes
#include <iostream>
#include <boost/python.hpp>
#include "Publisher.hh"

//#include "PythonPublisher.mpp"
#define prefix_
///////////////////////////////cc.p////////////////////////////////////////

namespace py = boost::python;

struct pykit::PythonPublisher::Impl
{
    boost::python::dict mainNamespace;
    Publisher * pythonPublisher;
};

#define PYTHON_PREPARE_IMPORT(module) \
    PyImport_AppendInittab(const_cast<char*>(#module), init ## module)

namespace {

    struct PublisherPyWrapper
        : public pykit::Publisher, py::wrapper<pykit::Publisher>
    {
        virtual void publish(pykit::Request & request)
            { get_override("publish")(request); }
    };

    namespace pyconvert {

        struct QString_PyUnicode
        {
            static PyObject * convert(QString const & s)
                {
                    std::wstring ws (s.toStdWString());
                    return PyUnicode_FromWideChar(ws.data(), ws.length());
                }
        };

        struct PyUnicode_QString
        {
            static void * convertible(PyObject * o)
                {
                    return PyUnicode_Check(o) ? o : 0;
                }

            static void construct(PyObject * o,
                                  py::converter::rvalue_from_python_stage1_data * data)
                {
                    unsigned length (PyUnicode_GetSize(o));
                    std::wstring ws (length, 0);
                    // Hmm ... I don't want to copy the stupid data TWICE but this
                    // breaks the standard ... who cares ?
                    PyUnicode_AsWideChar(reinterpret_cast<PyUnicodeObject *>(o),
                                         const_cast<wchar_t*>(ws.data()),
                                         length);
                    void * storage (((py::converter::rvalue_from_python_storage<QString>*)
                                     data)->storage.bytes);
                    *(new (storage) QString()) = QString::fromStdWString(ws);
                    data->convertible = storage;
                }

            PyUnicode_QString()
                {
                    py::converter::registry::push_back(
                        &convertible,
                        &construct,
                        py::type_id<QString>());
                }
        };

        struct QByteArray_PyString
        {
            static PyObject * convert(QByteArray const & s)
                {
                    return boost::python::incref(boost::python::object(s.constData()).ptr());
                }
        };

        struct PyString_QByteArray
        {
            static void * convertible(PyObject * o)
                {
                    return PyString_Check(o) ? o : 0;
                }

            static void construct(PyObject * o,
                                  py::converter::rvalue_from_python_stage1_data * data)
                {
                    unsigned length (PyString_Size(o));
                    const char * value (PyString_AsString(o));
                    void * storage (((py::converter::rvalue_from_python_storage<QByteArray>*)
                                     data)->storage.bytes);
                    new (storage) QByteArray(value,length);
                    data->convertible = storage;
                }

            PyString_QByteArray()
                {
                    py::converter::registry::push_back(
                        &convertible,
                        &construct,
                        py::type_id<QByteArray>());
                }
        };

    }

    QString QUrl_toString_noargs(QUrl const & url)
    { return url.toString(); }

}

#define MEMFNP(ret, cls, nam, arg) static_cast<ret (cls::*)arg>(&cls::nam)

BOOST_PYTHON_MODULE(_qt)
{
    py::to_python_converter<QString, pyconvert::QString_PyUnicode>();
    pyconvert::PyUnicode_QString register_PyUnicode_QString;

    py::to_python_converter<QByteArray, pyconvert::QByteArray_PyString>();
    pyconvert::PyString_QByteArray register_PyString_QByteArray;

    // Missing converters:
    //    QPair <-> tuple
    //    QList <-> vector

        // QUrl ()
    py::class_<QUrl>("QUrl", py::init<>())
        // QUrl ( const QString & url )
        .def(py::init<QString const &>())
        // QUrl ( const QUrl & other )
        .def(py::init<QUrl const &>())
        // QUrl ( const QString & url, ParsingMode parsingMode )
        // .def(init<QString const &, ParsingMode>())
        // ~QUrl ()
        // void   addEncodedQueryItem ( const QByteArray & key, const QByteArray & value )
        .def("addEncodedQueryItem",
             MEMFNP(void, QUrl, addEncodedQueryItem,
                    ( const QByteArray & key, const QByteArray & value )))
        // void   addQueryItem ( const QString & key, const QString & value )
        .def("addQueryItem",
             MEMFNP(void, QUrl, addQueryItem,
                    ( const QString & key, const QString & value )))
        // QList<QByteArray>   allEncodedQueryItemValues ( const QByteArray & key ) const
        // .def("allEncodedQueryItemValues",
        //      MEMFNP(QList<QByteArray>, QUrl, allEncodedQueryItemValues,
        //             ( const QByteArray & key ) const))
        // QStringList   allQueryItemValues ( const QString & key ) const
        // .def("allQueryItemValues",
        //      MEMFNP(QStringList, QUrl, allQueryItemValues,
        //             ( const QString & key ) const))
        // QString   authority () const
        .def("authority",
             MEMFNP(QString, QUrl, authority,
                    () const))
        // void   clear ()
        .def("clear",
             MEMFNP(void, QUrl, clear,
                    ()))
        // QByteArray   encodedFragment () const
        .def("encodedFragment",
             MEMFNP(QByteArray, QUrl, encodedFragment,
                    () const))
        // QByteArray   encodedHost () const
        .def("encodedHost",
             MEMFNP(QByteArray, QUrl, encodedHost,
                    () const))
        // QByteArray   encodedPassword () const
        .def("encodedPassword",
             MEMFNP(QByteArray, QUrl, encodedPassword,
                    () const))
        // QByteArray   encodedPath () const
        .def("encodedPath",
             MEMFNP(QByteArray, QUrl, encodedPath,
                    () const))
        // QByteArray   encodedQuery () const
        .def("encodedQuery",
             MEMFNP(QByteArray, QUrl, encodedQuery,
                    () const))
        // QByteArray   encodedQueryItemValue ( const QByteArray & key ) const
        .def("encodedQueryItemValue",
             MEMFNP(QByteArray, QUrl, encodedQueryItemValue,
                    ( const QByteArray & key ) const))
        // QList<QPair<QByteArray, QByteArray> >   encodedQueryItems () const
        // .def("encodedQueryItems",
        //      MEMFNP(QList<QPair<QByteArray, QByteArray> >, QUrl, encodedQueryItems,
        //             () const))
        // QByteArray   encodedUserName () const
        .def("encodedUserName",
             MEMFNP(QByteArray, QUrl, encodedUserName,
                    () const))
        // QString   errorString () const
        .def("errorString",
             MEMFNP(QString, QUrl, errorString,
                    () const))
        // QString   fragment () const
        .def("fragment",
             MEMFNP(QString, QUrl, fragment,
                    () const))
        // bool   hasEncodedQueryItem ( const QByteArray & key ) const
        .def("hasEncodedQueryItem",
             MEMFNP(bool, QUrl, hasEncodedQueryItem,
                    ( const QByteArray & key ) const))
        // bool   hasFragment () const
        .def("hasFragment",
             MEMFNP(bool, QUrl, hasFragment,
                    () const))
        // bool   hasQuery () const
        .def("hasQuery",
             MEMFNP(bool, QUrl, hasQuery,
                    () const))
        // bool   hasQueryItem ( const QString & key ) const
        .def("hasQueryItem",
             MEMFNP(bool, QUrl, hasQueryItem,
                    ( const QString & key ) const))
        // QString   host () const
        .def("host",
             MEMFNP(QString, QUrl, host,
                    () const))
        // bool   isEmpty () const
        .def("isEmpty",
             MEMFNP(bool, QUrl, isEmpty,
                    () const))
        // bool   isParentOf ( const QUrl & childUrl ) const
        .def("isParentOf",
             MEMFNP(bool, QUrl, isParentOf,
                    ( const QUrl & childUrl ) const))
        // bool   isRelative () const
        .def("isRelative",
             MEMFNP(bool, QUrl, isRelative,
                    () const))
        // bool   isValid () const
        .def("isValid",
             MEMFNP(bool, QUrl, isValid,
                    () const))
        // QString   password () const
        .def("password",
             MEMFNP(QString, QUrl, password,
                    () const))
        // QString   path () const
        .def("path",
             MEMFNP(QString, QUrl, path,
                    () const))
        // int   port () const
        .def("port",
             MEMFNP(int, QUrl, port,
                    () const))
        // int   port ( int defaultPort ) const
        .def("port",
             MEMFNP(int, QUrl, port,
                    ( int defaultPort ) const))
        // QString   queryItemValue ( const QString & key ) const
        .def("queryItemValue",
             MEMFNP(QString, QUrl, queryItemValue,
                    ( const QString & key ) const))
        // QList<QPair<QString, QString> >   queryItems () const
        // .def("queryItems",
        //      MEMFNP(QList<QPair<QString, QString> >, QUrl, queryItems,
        //             () const))
        // char   queryPairDelimiter () const
        .def("queryPairDelimiter",
             MEMFNP(char, QUrl, queryPairDelimiter,
                    () const))
        // char   queryValueDelimiter () const
        .def("queryValueDelimiter",
             MEMFNP(char, QUrl, queryValueDelimiter,
                    () const))
        // void   removeAllEncodedQueryItems ( const QByteArray & key )
        .def("removeAllEncodedQueryItems",
             MEMFNP(void, QUrl, removeAllEncodedQueryItems,
                    ( const QByteArray & key )))
        // void   removeAllQueryItems ( const QString & key )
        .def("removeAllQueryItems",
             MEMFNP(void, QUrl, removeAllQueryItems,
                    ( const QString & key )))
        // void   removeEncodedQueryItem ( const QByteArray & key )
        .def("removeEncodedQueryItem",
             MEMFNP(void, QUrl, removeEncodedQueryItem,
                    ( const QByteArray & key )))
        // void   removeQueryItem ( const QString & key )
        .def("removeQueryItem",
             MEMFNP(void, QUrl, removeQueryItem,
                    ( const QString & key )))
        // QUrl   resolved ( const QUrl & relative ) const
        .def("resolved",
             MEMFNP(QUrl, QUrl, resolved,
                    ( const QUrl & relative ) const))
        // QString   scheme () const
        .def("scheme",
             MEMFNP(QString, QUrl, scheme,
                    () const))
        // void   setAuthority ( const QString & authority )
        .def("setAuthority",
             MEMFNP(void, QUrl, setAuthority,
                    ( const QString & authority )))
        // void   setEncodedFragment ( const QByteArray & fragment )
        .def("setEncodedFragment",
             MEMFNP(void, QUrl, setEncodedFragment,
                    ( const QByteArray & fragment )))
        // void   setEncodedHost ( const QByteArray & host )
        .def("setEncodedHost",
             MEMFNP(void, QUrl, setEncodedHost,
                    ( const QByteArray & host )))
        // void   setEncodedPassword ( const QByteArray & password )
        .def("setEncodedPassword",
             MEMFNP(void, QUrl, setEncodedPassword,
                    ( const QByteArray & password )))
        // void   setEncodedPath ( const QByteArray & path )
        .def("setEncodedPath",
             MEMFNP(void, QUrl, setEncodedPath,
                    ( const QByteArray & path )))
        // void   setEncodedQuery ( const QByteArray & query )
        .def("setEncodedQuery",
             MEMFNP(void, QUrl, setEncodedQuery,
                    ( const QByteArray & query )))
        // void   setEncodedQueryItems ( const QList<QPair<QByteArray, QByteArray> > & query )
        // .def("setEncodedQueryItems",
        //      MEMFNP(void, QUrl, setEncodedQueryItems,
        //             ( const QList<QPair<QByteArray, QByteArray> > & query )))
        // void   setEncodedUrl ( const QByteArray & encodedUrl )
        .def("setEncodedUrl",
             MEMFNP(void, QUrl, setEncodedUrl,
                    ( const QByteArray & encodedUrl )))
        // void   setEncodedUrl ( const QByteArray & encodedUrl, ParsingMode parsingMode )
        // .def("setEncodedUrl",
        //      MEMFNP(void, QUrl, setEncodedUrl,
        //             ( const QByteArray & encodedUrl, ParsingMode parsingMode )))
        // void   setEncodedUserName ( const QByteArray & userName )
        .def("setEncodedUserName",
             MEMFNP(void, QUrl, setEncodedUserName,
                    ( const QByteArray & userName )))
        // void   setFragment ( const QString & fragment )
        .def("setFragment",
             MEMFNP(void, QUrl, setFragment,
                    ( const QString & fragment )))
        // void   setHost ( const QString & host )
        .def("setHost",
             MEMFNP(void, QUrl, setHost,
                    ( const QString & host )))
        // void   setPassword ( const QString & password )
        .def("setPassword",
             MEMFNP(void, QUrl, setPassword,
                    ( const QString & password )))
        // void   setPath ( const QString & path )
        .def("setPath",
             MEMFNP(void, QUrl, setPath,
                    ( const QString & path )))
        // void   setPort ( int port )
        .def("setPort",
             MEMFNP(void, QUrl, setPort,
                    ( int port )))
        // void   setQueryDelimiters ( char valueDelimiter, char pairDelimiter )
        .def("setQueryDelimiters",
             MEMFNP(void, QUrl, setQueryDelimiters,
                    ( char valueDelimiter, char pairDelimiter )))
        // void   setQueryItems ( const QList<QPair<QString, QString> > & query )
        // .def("setQueryItems",
        //      MEMFNP(void, QUrl, setQueryItems,
        //             ( const QList<QPair<QString, QString> > & query )))
        // void   setScheme ( const QString & scheme )
        .def("setScheme",
             MEMFNP(void, QUrl, setScheme,
                    ( const QString & scheme )))
        // void   setUrl ( const QString & url )
        .def("setUrl",
             MEMFNP(void, QUrl, setUrl,
                    ( const QString & url )))
        // void   setUrl ( const QString & url, ParsingMode parsingMode )
        // .def("setUrl",
        //      MEMFNP(void, QUrl, setUrl,
        //             ( const QString & url, ParsingMode parsingMode )))
        // void   setUserInfo ( const QString & userInfo )
        .def("setUserInfo",
             MEMFNP(void, QUrl, setUserInfo,
                    ( const QString & userInfo )))
        // void   setUserName ( const QString & userName )
        .def("setUserName",
             MEMFNP(void, QUrl, setUserName,
                    ( const QString & userName )))
        // QByteArray   toEncoded ( FormattingOptions options = None ) const
        // .def("toEncoded",
        //      MEMFNP(QByteArray, QUrl, toEncoded,
        //             ( FormattingOptions options = None ) const))
        // QString   toLocalFile () const
        .def("toLocalFile",
             MEMFNP(QString, QUrl, toLocalFile,
                    () const))
        // QString   toString ( FormattingOptions options = None ) const
        .def("__unicode__", &QUrl_toString_noargs)
        // QString   userInfo () const
        .def("userInfo",
             MEMFNP(QString, QUrl, userInfo,
                    () const))
        // QString   userName () const
        .def("userName",
             MEMFNP(QString, QUrl, userName,
                    () const))

        // bool   operator!= ( const QUrl & url ) const
        // QUrl &   operator= ( const QUrl & url )
        // QUrl &   operator= ( const QString & url )
        // bool   operator== ( const QUrl & url ) const
        ;
}

BOOST_PYTHON_MODULE(_pykit)
{
    py::class_<pykit::Request>("Request", py::no_init)
        .def("write", &pykit::Request::write)
        .def("reset", &pykit::Request::reset)
        .def("setContentType", &pykit::Request::setContentType)
        .def("setLocation", &pykit::Request::setLocation)
        .def("setRawHeader", &pykit::Request::setHeader)
        .def("setStatusCode", &pykit::Request::setStatusCode)
        .def("url", &pykit::Request::url)
        .def("postData", &pykit::Request::postData)
        .def("operation", &pykit::Request::operation)
        .def("postContentType", &pykit::Request::postContentType)
        ;

    py::class_<PublisherPyWrapper, boost::noncopyable>("Publisher")
        .def("publish", py::pure_virtual(&pykit::Publisher::publish))
        ;
}

prefix_ pykit::PythonPublisher::PythonPublisher()
    : impl_ (new Impl)
{
    try {
        PYTHON_PREPARE_IMPORT(_pykit);
        PYTHON_PREPARE_IMPORT(_qt);
        Py_Initialize();
        py::object initModule = py::import("init");
        impl_->mainNamespace = py::extract<py::dict>(initModule.attr("__dict__"));
        impl_->pythonPublisher = py::extract<Publisher*>(impl_->mainNamespace["initialize"]());
    }
    catch (boost::python::error_already_set & ex) {
        PyErr_Print();
        throw;
    }
}

prefix_ pykit::PythonPublisher::~PythonPublisher()
{}

prefix_ void pykit::PythonPublisher::publish(Request & request)
{
    try {
        impl_->pythonPublisher->publish(request);
    }
    catch (py::error_already_set & ex) {
        PyErr_Print();
    }
}

///////////////////////////////cc.e////////////////////////////////////////
#undef prefix_
//#include "PythonPublisher.mpp"


// Local Variables:
// mode: c++
// fill-column: 100
// comment-column: 40
// c-file-style: "j32"
// indent-tabs-mode: nil
// ispell-local-dictionary: "american"
// compile-command: "scons -U"
// End:
